#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "syntax.h"
#include "area.h"
#include "value.h"

/*****************************************************************
 * value_t
 */

value_t value_destroy(value_t value)
{
  int i;

  if (value != NULL) {
    switch (value->type) {
    case C_TYPE_NUMBER:
      break;
    case C_TYPE_STRING:
      if (value->obj.string.s != NULL)
	free(value->obj.string.s);
      break;
    case C_TYPE_ARRAY:
      if (value->obj.array.values)
	objlist_destroy(value->obj.array.values);
      break;
    case C_TYPE_VARIABLE:
     break;
    case C_TYPE_FUNCTION:
      for (i = 0; i < value->obj.function.args.number; i++) {
	if (value->obj.function.args.arg[i])
	  symbol_destroy(value->obj.function.args.arg[i]);
      }
      break;
    case C_TYPE_BLOCK:
      syntax_destroy(value->obj.block.syntax);
      break;
    default:
      for (i = 0; i < VALUE_ARG_NUM; i++) {
	if (value->obj.op.arg[i])
	  symbol_destroy(value->obj.op.arg[i]);
      }
      break;
    }

    memset(value, 0, sizeof(*value));
    free(value);
  }

  return NULL;
}

value_t value_create(c_type_t type)
{
  value_t value;

  value = malloc(sizeof(*value));
  if (value == NULL)
    goto err;
  memset(value, 0, sizeof(*value));

  value->type = type;

  return value;

err:
  if (value != NULL)
    value_destroy(value);
  return NULL;
}

static void value_print_array(value_t value, int indent)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(value->obj.array.values);
       !objlist_is_end(value->obj.array.values, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    symbol_print(symbol, indent);
  }
}

void value_print_simple(value_t value, int indent)
{
  int i;

  indent_print(indent);
  printf("VALUE:  ");

  switch (value->type) {
  case C_TYPE_NUMBER:
    printf("%d (%s)\n", value->obj.number, type_get_word(value->type));
    break;
  case C_TYPE_STRING:
    printf("%s (%s)\n", value->obj.string.s, type_get_word(value->type));
    break;
  case C_TYPE_ARRAY:
    printf("%d (%s)\n", value->obj.array.number, type_get_word(value->type));
    break;
  case C_TYPE_LABEL:
    symbol_print_name(value->obj.label.symbol);
    printf(" (%s:0x%x)\n", type_get_word(value->type), value->obj.label.symbol->id);
    break;
  case C_TYPE_VARIABLE:
    symbol_print_name(value->obj.variable.symbol);
    printf(" (%s:0x%x)\n", type_get_word(value->type), value->obj.variable.symbol->id);
    break;
  case C_TYPE_FUNCTION:
    symbol_print_name(value->obj.function.symbol);
    printf(" (%s:0x%x) (", type_get_word(value->type), value->obj.function.symbol->id);
    for (i = 0; i < value->obj.function.args.number; i++) {
      printf(" ");
      symbol_print_name(value->obj.function.args.arg[i]);
    }
    printf(" )\n");
    break;
  case C_TYPE_ARG:
    printf("%d (%s)\n", value->obj.arg.number, type_get_word(value->type));
    break;
  case C_TYPE_BLOCK:
    printf("%s\n", type_get_word(value->type));
    break;
  default:
    printf("%s (", type_get_word(value->type));
    for (i = 0; i < VALUE_ARG_NUM; i++) {
      if (value->obj.op.arg[i] != NULL) {
	printf(" ");
	symbol_print_name(value->obj.op.arg[i]);
      }
    }
    printf(" )\n");
    break;
  }
}

void value_print(value_t value, int indent)
{
  int i;

  value_print_simple(value, indent);

  switch (value->type) {
  case C_TYPE_NUMBER:
  case C_TYPE_STRING:
    break;
  case C_TYPE_ARRAY:
    if (value->obj.array.values != NULL)
      value_print_array(value, indent + 1);
    break;
  case C_TYPE_LABEL:
  case C_TYPE_VARIABLE:
    break;
  case C_TYPE_FUNCTION:
    indent++;
    symbol_print(value->obj.function.symbol, indent);
    indent_print(indent);
    printf("ARGS\n");
    for (i = 0; i < value->obj.function.args.number; i++) {
      symbol_print_simple(value->obj.function.args.arg[i], indent + 1);
    }
    break;
  case C_TYPE_ARG:
    break;
  case C_TYPE_BLOCK:
    if (value->obj.block.syntax != NULL)
      syntax_print(value->obj.block.syntax, indent + 1);
    break;
  default:
    break;
  }
}

void value_print_args(value_t value, int indent)
{
  int i;

  value_print_simple(value, indent);

  switch (value->type) {
  case C_TYPE_NUMBER:
  case C_TYPE_STRING:
  case C_TYPE_ARRAY:
  case C_TYPE_LABEL:
  case C_TYPE_VARIABLE:
    break;
  case C_TYPE_FUNCTION:
    for (i = 0; i < value->obj.function.args.number; i++) {
      symbol_print_value_simple(value->obj.function.args.arg[i], indent + 1);
    }
    break;
  case C_TYPE_ARG:
  case C_TYPE_BLOCK:
    break;
  default:
    for (i = 0; i < VALUE_ARG_NUM; i++) {
      if (value->obj.op.arg[i] != NULL)
	symbol_print_value_simple(value->obj.op.arg[i], indent + 1);
    }
    break;
  }
}

/*****************************************************************
 * library
 */

static int is_const_variable(symbol_t symbol)
{
  if ((symbol->type == C_TYPE_VARIABLE) &&
      (symbol->flags & SYMBOL_FLAG_CONSTVALUE))
    return 1;
  return 0;
}

static int is_const_number(symbol_t symbol)
{
  if (is_const_variable(symbol) &&
      (symbol->value != NULL) &&
      (symbol->value->type == C_TYPE_NUMBER))
    return 1;
  return 0;
}

static int is_writable(symbol_t symbol)
{
  if ((symbol->type != C_TYPE_VARIABLE) || is_const_variable(symbol))
    return 0;
  return 1;
}

static int is_pointer(model_t model)
{
  return (model_get_modifier_number(model) > 0) ? 1 : 0;
}

static int model_level(model_t model)
{
  int r = 0;
  if (is_pointer(model))
    return 8;
  switch (model->type) {
  case C_TYPE_CHAR:  r = 0; break;
  case C_TYPE_SHORT: r = 2; break;
  case C_TYPE_INT:   r = 4; break;
  case C_TYPE_LONG:  r = 6; break;
  default: r = 0; break;
  }
  return r + ((model->flags & MODEL_FLAG_UNSIGNED) ? 1 : 0);
}

static symbol_t make_cast(symbol_t symbol, model_t model,
			  objlist_t symbols,
			  area_t static_area, area_t stack_area)
{
  symbol_t s;

  symbol = symbol_locate(symbol, symbols, static_area, stack_area);

  s = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE,
		    SYMBOL_NAME_AUTO, NULL, 0, C_TYPE_CAST);
  if (s == NULL)
    return NULL;
  if (symbols)
    objlist_insert_obj_tail(symbols, s, symbol_free);
  s->model = model;
  s->value->obj.op.arg[0] = symbol;

  return s;
}

static symbol_t change_model(symbol_t symbol, model_t model,
			     objlist_t symbols,
			     area_t static_area, area_t stack_area)
{
  symbol_t s;

  if (model_level(symbol->model) == model_level(model))
    return symbol;

  s = make_cast(symbol, model_copy(model), symbols, static_area, stack_area);
  if (s == NULL)
    ERREXIT(1);

  return s;
}

static symbol_t change_model_type(symbol_t symbol, c_type_t type,
				  objlist_t symbols,
				  area_t static_area, area_t stack_area)
{
  symbol_t s;

  if (symbol->model->type == type)
    return symbol;

  s = make_cast(symbol, model_create(type), symbols, static_area, stack_area);
  if (s == NULL)
    ERREXIT(1);

  return s;
}

static int matching_model(syntax_t syntax0, syntax_t syntax1,
			  objlist_t symbols,
			  area_t static_area, area_t stack_area)
{
  int level0, level1;

  if ((syntax0->obj.symbol->model == NULL) ||
      (syntax1->obj.symbol->model == NULL))
    return -1;

  level0 = model_level(syntax0->obj.symbol->model);
  level1 = model_level(syntax1->obj.symbol->model);

  if (level0 != level1) {
    if (level0 > level1)
      syntax1->obj.symbol = change_model(syntax1->obj.symbol,
					 syntax0->obj.symbol->model,
					 symbols, static_area, stack_area);
    else
      syntax0->obj.symbol = change_model(syntax0->obj.symbol,
					 syntax1->obj.symbol->model,
					 symbols, static_area, stack_area);
  }

  return 0;
}

static symbol_t _value_read(objlist_t words,
			    objlist_t defines_list, objlist_t symbols,
			    area_t static_area, area_t stack_area);

static int op1(c_type_t type, syntax_t syntax,
	       objlist_t words, objlist_t defines_list, objlist_t symbols,
	       area_t static_area, area_t stack_area)
{
  objentry_t pentry, nentry;
  syntax_t psyntax = NULL, nsyntax = NULL;
  symbol_t symbol, s = NULL;

  saveline(syntax->line);

  pentry = objentry_get_prev(&syntax->entry);
  nentry = objentry_get_next(&syntax->entry);

  if (!objlist_is_end(words, pentry))
    psyntax = objentry_get_syntax(pentry);
  if (!objlist_is_end(words, nentry))
    nsyntax = objentry_get_syntax(nentry);

  switch (type) {
  case C_TYPE_AND: /* &a */
  case C_TYPE_MUL: /* *p */
  case C_TYPE_ADD: /* +i */
  case C_TYPE_SUB: /* -i */
    if ((psyntax != NULL) &&
	(((psyntax->type == C_TYPE_SYMBOL) &&
	  ((psyntax->obj.symbol->value == NULL) ||
	   (psyntax->obj.symbol->value->type != C_TYPE_CAST))) ||
	 (psyntax->type == C_TYPE_INC) || (psyntax->type == C_TYPE_DEC))) {
      return 0;
    }
    /* fall through */

  case C_TYPE_INV: /* ~i */
  case C_TYPE_NOT: /* !i */
    objlist_extract(words, nentry);
    s = nsyntax->obj.symbol;

    if (type == C_TYPE_AND) {
      if (s->type == C_TYPE_LABEL) {
	syntax->type = C_TYPE_SYMBOL;
	syntax->obj.symbol = s;
	return 0;
      }
    }

    if (type == C_TYPE_MUL) {
      if ((s->model != NULL) && (s->model->type == C_TYPE_FUNCTION) &&
	  (s->model->modifiers.number == 1)) {
	syntax->type = C_TYPE_SYMBOL;
	syntax->obj.symbol = s;
	return 0;
      }
    }

    if (type != C_TYPE_AND)
      nsyntax->obj.symbol = symbol_locate(nsyntax->obj.symbol, symbols,
					  static_area, stack_area);

    switch (type) {
    case C_TYPE_AND: type = C_TYPE_ADDRESS; break;
    case C_TYPE_MUL: type = C_TYPE_POINTER; break;
    case C_TYPE_ADD: type = C_TYPE_PLUS;    break;
    case C_TYPE_SUB: type = C_TYPE_MINUS;   break;
    default: break;
    }

    syntax->type = C_TYPE_SYMBOL;
    symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE,
			   SYMBOL_NAME_AUTO, NULL, 0, type);
    if (symbol == NULL)
      ERREXIT(1);
    if (symbols)
      objlist_insert_obj_tail(symbols, symbol, symbol_free);
    syntax->obj.symbol = symbol;
    symbol->model = model_copy(s->model);

    if (type == C_TYPE_POINTER) {
      if (model_del_modifier(symbol->model) < 0)
	ERREXIT(1);
      model_array_to_pointer(symbol->model);
      symbol->location.type = C_TYPE_POINTER;
      symbol->location.base.symbol = s;
      symbol->location.offset = 0;
      symbol->flags |= SYMBOL_FLAG_LOCATED;
    } else {
      if (type == C_TYPE_ADDRESS) {
	if (model_add_modifier_pointer(symbol->model) < 0)
	  ERREXIT(1);
      }
    }
    break;

  case C_TYPE_INC: /* ++ */
  case C_TYPE_DEC: /* -- */
    if ((nsyntax != NULL) && (nsyntax->type == C_TYPE_SYMBOL)) { /* ++i/--i */
      if (!is_writable(nsyntax->obj.symbol))
	ERREXIT(1);
      objlist_extract(words, nentry);
      s = nsyntax->obj.symbol;
    } else if ((psyntax != NULL) && (psyntax->type == C_TYPE_SYMBOL)) { /* i++/i-- */
      if (!is_writable(psyntax->obj.symbol))
	ERREXIT(1);
      objlist_extract(words, pentry);
      s = psyntax->obj.symbol;
    } else {
      ERREXIT(1);
    }

    syntax->type = C_TYPE_SYMBOL;
    symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE,
			   SYMBOL_NAME_AUTO, NULL, 0, type);
    if (symbol == NULL)
      ERREXIT(1);
    if (symbols)
      objlist_insert_obj_tail(symbols, symbol, symbol_free);
    syntax->obj.symbol = symbol;
    symbol->model = model_copy(s->model);
    break;

  default:
    ERREXIT(1);
    return -1;
  }

  symbol->value->obj.op.arg[0] = s;

  return 0;
}

static int op2_const(c_type_t type, symbol_t symbol,
		     symbol_t psymbol, symbol_t nsymbol)
{
  int n, pn, nn;

  if (!is_const_number(psymbol) || !is_const_number(nsymbol))
    return 0;

  pn = psymbol->value->obj.number;
  nn = nsymbol->value->obj.number;

  if ((type == C_TYPE_DIV) || (type == C_TYPE_MOD)) {
    if (nn == 0)
      ERREXIT(1);
  }

  switch (type) {
  case C_TYPE_MUL:    n = pn *  nn; break;
  case C_TYPE_DIV:    n = pn /  nn; break;
  case C_TYPE_MOD:    n = pn %  nn; break;
  case C_TYPE_ADD:    n = pn +  nn; break;
  case C_TYPE_SUB:    n = pn -  nn; break;
  case C_TYPE_RSHIFT: n = pn >> nn; break;
  case C_TYPE_LSHIFT: n = pn << nn; break;
  case C_TYPE_GTEQ:   n = pn >= nn; break;
  case C_TYPE_LTEQ:   n = pn <= nn; break;
  case C_TYPE_GT:     n = pn >  nn; break;
  case C_TYPE_LT:     n = pn <  nn; break;
  case C_TYPE_EQEQ:   n = pn == nn; break;
  case C_TYPE_NEQ:    n = pn != nn; break;
  case C_TYPE_AND:    n = pn &  nn; break;
  case C_TYPE_XOR:    n = pn ^  nn; break;
  case C_TYPE_OR:     n = pn |  nn; break;
  case C_TYPE_ANDAND: n = pn && nn; break;
  case C_TYPE_OROR:   n = pn || nn; break;
  default:
    return 0;
  }

  symbol->value->type = C_TYPE_NUMBER;
  symbol->value->obj.number = n;
  symbol->model = model_create(C_TYPE_INT);
  symbol->flags |= SYMBOL_FLAG_CONSTVALUE;

  return 1;
}

static int op2(c_type_t type, syntax_t syntax,
	       objlist_t words, objlist_t defines_list, objlist_t symbols,
	       area_t static_area, area_t stack_area)
{
  syntax_t psyntax, nsyntax;
  symbol_t symbol;

  saveline(syntax->line);

  psyntax = objlist_extract_syntax_prev(words, syntax);
  nsyntax = objlist_extract_syntax_next(words, syntax);

  if ((psyntax == NULL) || (nsyntax == NULL) ||
      (psyntax->type != C_TYPE_SYMBOL) || (nsyntax->type != C_TYPE_SYMBOL))
    ERREXIT(1);

  symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE, SYMBOL_NAME_AUTO,
			 NULL, 0, type);
  if (symbol == NULL)
    ERREXIT(1);
  if (symbols)
    objlist_insert_obj_tail(symbols, symbol, symbol_free);

  syntax->type = C_TYPE_SYMBOL;
  syntax->obj.symbol = symbol;

  if (op2_const(type, symbol, psyntax->obj.symbol, nsyntax->obj.symbol))
    return 0;

  switch (type) {
  case C_TYPE_EQ:
  case C_TYPE_ADDEQ:
  case C_TYPE_SUBEQ:
  case C_TYPE_MULEQ:
  case C_TYPE_DIVEQ:
  case C_TYPE_MODEQ:
  case C_TYPE_ANDEQ:
  case C_TYPE_XOREQ:
  case C_TYPE_OREQ:
  case C_TYPE_RSHIFTEQ:
  case C_TYPE_LSHIFTEQ:
    if (!is_writable(psyntax->obj.symbol))
      ERREXIT(1);
    break;

  default:
    break;
  }

  switch (type) {
  case C_TYPE_ADD:
  case C_TYPE_SUB:
    if (is_pointer(psyntax->obj.symbol->model) &&
	is_pointer(nsyntax->obj.symbol->model)) {
      if (type == C_TYPE_SUB) {
	symbol->model = model_create(C_TYPE_INT);
	break;
      } else
	ERREXIT(1);
    } else if (is_pointer(psyntax->obj.symbol->model)) {
      nsyntax->obj.symbol = change_model_type(nsyntax->obj.symbol, C_TYPE_INT, symbols, static_area, stack_area);
      symbol->model = model_copy(psyntax->obj.symbol->model);
      break;
    } else if (is_pointer(nsyntax->obj.symbol->model)) {
      psyntax->obj.symbol = change_model_type(psyntax->obj.symbol, C_TYPE_INT, symbols, static_area, stack_area);
      symbol->model = model_copy(nsyntax->obj.symbol->model);
      break;
    }
    /* fall through */

  case C_TYPE_MUL:
  case C_TYPE_DIV:
  case C_TYPE_MOD:
  case C_TYPE_AND:
  case C_TYPE_XOR:
  case C_TYPE_OR:
  case C_TYPE_RSHIFT:
  case C_TYPE_LSHIFT:
    if (is_pointer(psyntax->obj.symbol->model) ||
	is_pointer(nsyntax->obj.symbol->model))
      ERREXIT(1);
    if ((type == C_TYPE_RSHIFT) || (type == C_TYPE_LSHIFT)) {
      nsyntax->obj.symbol = change_model_type(nsyntax->obj.symbol, C_TYPE_INT, symbols, static_area, stack_area);
    } else {
      if (matching_model(psyntax, nsyntax, symbols, static_area, stack_area) < 0)
	ERREXIT(1);
    }
    symbol->model = model_copy(psyntax->obj.symbol->model);
    break;

  case C_TYPE_GTEQ:
  case C_TYPE_LTEQ:
  case C_TYPE_GT:
  case C_TYPE_LT:
  case C_TYPE_EQEQ:
  case C_TYPE_NEQ:
    if (matching_model(psyntax, nsyntax, symbols, static_area, stack_area) < 0)
      ERREXIT(1);
    symbol->model = model_create(C_TYPE_INT);
    break;

  case C_TYPE_ANDAND:
  case C_TYPE_OROR:
    symbol->model = model_create(C_TYPE_INT);
    break;

  case C_TYPE_EQ:
  case C_TYPE_ADDEQ:
  case C_TYPE_SUBEQ:
  case C_TYPE_MULEQ:
  case C_TYPE_DIVEQ:
  case C_TYPE_MODEQ:
  case C_TYPE_ANDEQ:
  case C_TYPE_XOREQ:
  case C_TYPE_OREQ:
  case C_TYPE_RSHIFTEQ:
  case C_TYPE_LSHIFTEQ:
    nsyntax->obj.symbol = change_model(nsyntax->obj.symbol, psyntax->obj.symbol->model, symbols, static_area, stack_area);
    symbol->model = model_copy(psyntax->obj.symbol->model);
    break;

  case C_TYPE_COMMA:
    nsyntax->obj.symbol = change_model(nsyntax->obj.symbol, psyntax->obj.symbol->model, symbols, static_area, stack_area);
    symbol->model = model_copy(psyntax->obj.symbol->model);
    break;

  default:
    ERREXIT(1);
    break;
  }

  nsyntax->obj.symbol = symbol_locate(nsyntax->obj.symbol, symbols,
				      static_area, stack_area);

  switch (type) {
  case C_TYPE_ADD:
  case C_TYPE_SUB:
  case C_TYPE_MUL:
  case C_TYPE_DIV:
  case C_TYPE_MOD:
  case C_TYPE_AND:
  case C_TYPE_XOR:
  case C_TYPE_OR:
  case C_TYPE_RSHIFT:
  case C_TYPE_LSHIFT:
  case C_TYPE_GTEQ:
  case C_TYPE_LTEQ:
  case C_TYPE_GT:
  case C_TYPE_LT:
  case C_TYPE_EQEQ:
  case C_TYPE_NEQ:
  case C_TYPE_ANDAND:
  case C_TYPE_OROR:
  case C_TYPE_COMMA:
    psyntax->obj.symbol = symbol_locate(psyntax->obj.symbol, symbols,
					static_area, stack_area);
    break;

  default:
    break;
  }

  model_array_to_pointer(symbol->model);

  symbol->value->obj.op.arg[0] = psyntax->obj.symbol;
  symbol->value->obj.op.arg[1] = nsyntax->obj.symbol;

  return 0;
}

static int op3(c_type_t type, syntax_t syntax,
	       objlist_t words, objlist_t defines_list, objlist_t symbols,
	       area_t static_area, area_t stack_area)
{
  syntax_t syntax0, syntax1, syntax2;
  symbol_t symbol;

  saveline(syntax->line);

  syntax0 = objlist_extract_syntax_prev(words, syntax);
  syntax1 = objlist_extract_syntax_next(words, syntax);
  syntax2 = objlist_extract_syntax_next(words, syntax);
  if ((syntax0 == NULL) || (syntax1 == NULL) || (syntax2 == NULL) ||
      (syntax0->type != C_TYPE_SYMBOL) ||
      (syntax1->type != C_TYPE_SYMBOL) ||
      (syntax2->type != C_TYPE_SYMBOL))
    return -1;

  syntax->type = C_TYPE_SYMBOL;

  if (type == C_TYPE_QUESTION) {
    if (is_const_number(syntax0->obj.symbol)) {
      if (syntax0->obj.symbol->value->obj.number != 0) {
	syntax->obj.symbol = syntax1->obj.symbol;
      } else {
	syntax->obj.symbol = syntax2->obj.symbol;
      }
      return 0;
    } else {
      if (matching_model(syntax1, syntax2, symbols, static_area, stack_area) < 0)
	return -1;
    }
  }

  syntax0->obj.symbol = symbol_locate(syntax0->obj.symbol, symbols,
				      static_area, stack_area);
  syntax1->obj.symbol = symbol_locate(syntax1->obj.symbol, symbols,
				      static_area, stack_area);
  syntax2->obj.symbol = symbol_locate(syntax2->obj.symbol, symbols,
				      static_area, stack_area);

  symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE,
			 SYMBOL_NAME_AUTO, NULL, 0, type);
  if (symbol == NULL)
    ERREXIT(1);
  if (symbols)
    objlist_insert_obj_tail(symbols, symbol, symbol_free);

  syntax->obj.symbol = symbol;

  symbol->model = model_copy(syntax1->obj.symbol->model);
  symbol->value->obj.op.arg[0] = syntax0->obj.symbol;
  symbol->value->obj.op.arg[1] = syntax1->obj.symbol;
  symbol->value->obj.op.arg[2] = syntax2->obj.symbol;

  return 0;
}

static symbol_t read_array(objlist_t words,
			   objlist_t defines_list, objlist_t symbols,
			   area_t static_area, area_t stack_area)
{
  objentry_t entry;
  symbol_t symbol, s;

  symbol = symbol_create(C_TYPE_LABEL, C_TYPE_NONE, SYMBOL_NAME_AUTO,
			 NULL, 0, C_TYPE_ARRAY);
  if (symbols)
    objlist_insert_obj_tail(symbols, symbol, symbol_free);
  symbol->flags |= SYMBOL_FLAG_CONSTVALUE;

  symbol->value->obj.array.values = objlist_create(NULL);

  while (!objlist_is_empty(words)) {
    s = value_read(words, defines_list, symbols, static_area, stack_area,
		   NULL, C_TYPE_COMMA, NULL, 1, 0);
    if (s == NULL)
      ERREXIT(1);
    entry = objlist_insert_obj_tail(symbol->value->obj.array.values,
				    s, symbol_free);
    if (entry == NULL)
      ERREXIT(1);
    symbol->value->obj.array.number++;
  }

  return symbol;
}

static int read_function_args(value_t value, objlist_t words,
			      objlist_t defines_list, objlist_t symbols,
			      area_t static_area, area_t stack_area)
{
  symbol_t symbol;
  int argc = 0;

  while (!objlist_is_empty(words)) {
    if (argc == VALUE_FUNCTION_ARGS_NUM)
      ERREXIT(1);
    symbol = value_read(words, defines_list, symbols, static_area, stack_area,
			NULL, C_TYPE_COMMA, NULL, 1, 0);
    if (symbol == NULL)
      ERREXIT(1);
    symbol = symbol_locate(symbol, symbols, static_area, stack_area);
    value->obj.function.args.arg[argc++] = symbol;
    value->obj.function.args.number = argc;
  }

  return 0;
}

static symbol_t type_read(objlist_t words,
			  objlist_t defines_list, objlist_t symbols)
{
  objentry_t entry;
  syntax_t syntax;
  typedef_t _typedef;

  entry = objlist_get_head(words);
  if (objlist_is_end(words, entry))
    return NULL;
  syntax = objentry_get_syntax(entry);

  saveline(syntax->line);

  if (syntax->type == C_TYPE_WORD) {
    _typedef = defines_list_search_typedef_name(defines_list, syntax->obj.word);
    if (_typedef != NULL) {
      syntax->type = C_TYPE_TYPEDEF;
      syntax->obj.custom._typedef = _typedef;
    }
  }

  switch (syntax->type) {
  case C_TYPE_STRUCT:
  case C_TYPE_UNION:
  case C_TYPE_ENUM:
  case C_TYPE_SIGNED:
  case C_TYPE_UNSIGNED:
  case C_TYPE_SHORT:
  case C_TYPE_LONG:
  case C_TYPE_CHAR:
  case C_TYPE_INT:
  case C_TYPE_FLOAT:
  case C_TYPE_DOUBLE:
  case C_TYPE_VOID:
  case C_TYPE_TYPEDEF:
    return symbol_read(words, defines_list, symbols);
  default:
    break;
  }
  return NULL;
}

static symbol_t _value_read(objlist_t words,
			    objlist_t defines_list, objlist_t symbols,
			    area_t static_area, area_t stack_area)
{
  objentry_t entry, pentry, nentry;
  syntax_t syntax, s, psyntax, nsyntax;
  symbol_t symbol;
  objlist_t w;
  int count, size;

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    switch (syntax->type) {
    case C_TYPE_WORD:
      if (defines_list == NULL)
	ERREXIT(1);
      symbol = defines_list_search_symbol_name(defines_list, syntax->obj.word);
      if (symbol == NULL) {
	symbol = enum_list_search_symbol_defines(defines_list, syntax->obj.word);
      }
      if (symbol == NULL)
	ERREXIT(1);
      syntax->type = C_TYPE_SYMBOL;
      syntax->obj.symbol = symbol;
      break;

    case C_TYPE_SYMBOL:
      symbol = syntax->obj.symbol;
      if (!is_const_number(symbol)) {
	if (symbols)
	  objlist_insert_obj_tail(symbols, symbol, symbol_free);
      }
      break;

    case C_TYPE_BRACKET_IN:
      pentry = objentry_get_prev(entry);
      psyntax = !objlist_is_end(words, pentry) ? objentry_get_syntax(pentry) : NULL;
      if ((psyntax != NULL) &&
	  (psyntax->type == C_TYPE_SYMBOL) &&
	  (psyntax->obj.symbol->model->type == C_TYPE_FUNCTION) &&
	  (psyntax->obj.symbol->model->modifiers.number <= 1)) { /* function call */
#if 0
	if (psyntax->obj.symbol->model->modifiers.number == 0)
	  ERREXIT(1);
#endif
	psyntax->obj.symbol = symbol_locate(psyntax->obj.symbol, symbols,
					    static_area, stack_area);
	objlist_extract(words, pentry);
	symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE, SYMBOL_NAME_AUTO,
			       NULL, 0, C_TYPE_FUNCTION);
	if (symbol == NULL)
	  ERREXIT(1);
	if (symbols)
	  objlist_insert_obj_tail(symbols, symbol, symbol_free);
	symbol->model = model_copy(psyntax->obj.symbol->model->function.retval);
	symbol->value->obj.function.symbol = psyntax->obj.symbol;
	read_function_args(symbol->value, syntax->obj.words,
			   defines_list, symbols, static_area, stack_area);
	syntax->type = C_TYPE_SYMBOL;
	syntax->obj.symbol = symbol;
	break;
      }
      symbol = type_read(syntax->obj.words, defines_list, symbols);
      if (symbol != NULL) { /* cast */
	symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE,
			       SYMBOL_NAME_AUTO, symbol->model, 0, C_TYPE_CAST);
	if (symbol == NULL)
	  ERREXIT(1);
	if (symbols)
	  objlist_insert_obj_tail(symbols, symbol, symbol_free);
	syntax->type = C_TYPE_SYMBOL;
	syntax->obj.symbol = symbol;
	break;
      }
      symbol = _value_read(syntax->obj.words, defines_list, symbols,
			   static_area, stack_area);
      if (symbol == NULL)
	ERREXIT(1);
      syntax->type = C_TYPE_SYMBOL;
      syntax->obj.symbol = symbol;
      break;

    case C_TYPE_BBRACKET_IN:
      pentry = objentry_get_prev(entry);
      if (objlist_is_end(words, pentry))
	ERREXIT(1);
      psyntax = objentry_get_syntax(pentry);
      if (psyntax->type != C_TYPE_SYMBOL)
	ERREXIT(1);

      /* p[i] -> *(p + i) */

      s = syntax_create(C_TYPE_SYMBOL, syntax->line, 0, NULL);
      w = syntax->obj.words;
      syntax->obj.words = NULL;
      s->obj.symbol = _value_read(w, defines_list, symbols, static_area, stack_area);
      if (s->obj.symbol == NULL)
	ERREXIT(1);
      objlist_insert_next(words, entry, &s->entry);
      syntax->type = C_TYPE_ADD;
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);

      s = syntax_create(C_TYPE_SYMBOL, syntax->line, 0, NULL);
      s->obj.symbol = syntax->obj.symbol;
      syntax->obj.symbol = NULL;
      objlist_insert_next(words, entry, &s->entry);
      syntax->type = C_TYPE_MUL;
      if (op1(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
      break;

    case C_TYPE_MBRACKET_IN:
      symbol = read_array(syntax->obj.words, defines_list, symbols,
			  static_area, stack_area);
      if (symbol == NULL)
	ERREXIT(1);
      syntax->type = C_TYPE_SYMBOL;
      syntax->obj.symbol = symbol;
      break;

    case C_TYPE_ARROW:
      pentry = objentry_get_prev(entry);
      if (pentry == NULL)
	ERREXIT(1);
      psyntax = objentry_get_syntax(pentry);
      if (psyntax->type != C_TYPE_SYMBOL)
	ERREXIT(1);

      /* p->m -> (*p).m */
      s = syntax_create(C_TYPE_MUL, syntax->line, 0, NULL);
      if (s == NULL)
	ERREXIT(1);
      objlist_insert_prev(words, &psyntax->entry, &s->entry);
      if (op1(s->type, s, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
      syntax->type = C_TYPE_DOT;
      /* fall through */

    case C_TYPE_DOT:
      pentry = objlist_extract_prev(words, entry);
      nentry = objlist_extract_next(words, entry);
      if ((pentry == NULL) || (nentry == NULL))
	ERREXIT(1);
      psyntax = objentry_get_syntax(pentry);
      nsyntax = objentry_get_syntax(nentry);
      if (psyntax->type != C_TYPE_SYMBOL)
	ERREXIT(1);
      if (nsyntax->type != C_TYPE_WORD)
	ERREXIT(1);
      if (psyntax->obj.symbol->model->modifiers.number > 0)
	ERREXIT(1);

      syntax->obj.symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE, SYMBOL_NAME_AUTO,
					 NULL, 0, syntax->type);
      if (symbols)
	objlist_insert_obj_tail(symbols, syntax->obj.symbol, symbol_free);
      syntax->type = C_TYPE_SYMBOL;

      symbol = NULL;
      switch (psyntax->obj.symbol->model->type) {
      case C_TYPE_STRUCT:
	symbol = area_search_symbol(psyntax->obj.symbol->model->custom._struct->members, nsyntax->obj.word);
	break;
      case C_TYPE_UNION:
	symbol = area_search_symbol(psyntax->obj.symbol->model->custom._union->members, nsyntax->obj.word);
	break;
      default:
	ERREXIT(1);
      }
      if (symbol == NULL)
	ERREXIT(1);
      syntax->obj.symbol->model = model_copy(symbol->model);
      syntax->obj.symbol->value->obj.op.arg[0] = psyntax->obj.symbol;
      syntax->obj.symbol->value->obj.op.arg[1] = symbol;
      syntax->obj.symbol->location.type = C_TYPE_SYMBOL;
      syntax->obj.symbol->location.base.symbol = psyntax->obj.symbol;
      syntax->obj.symbol->location.offset = symbol->location.offset;
      syntax->obj.symbol->flags |= SYMBOL_FLAG_LOCATED;
      break;

    default:
      break;
    }
  }

  for (entry = objlist_get_tail(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_prev(entry)) { /* right to left */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    switch (syntax->type) {
    case C_TYPE_SIZEOF:
      nentry = objlist_extract_next(words, entry);
      if (nentry == NULL)
	ERREXIT(1);
      nsyntax = objentry_get_syntax(nentry);
      if (nsyntax->type != C_TYPE_SYMBOL)
	ERREXIT(1);
      syntax->type = C_TYPE_SYMBOL;
      symbol = symbol_create(C_TYPE_VARIABLE, C_TYPE_NONE, SYMBOL_NAME_AUTO,
			     NULL, 0, C_TYPE_NUMBER);
      if (symbols)
	objlist_insert_obj_tail(symbols, symbol, symbol_free);
      size = model_get_size(nsyntax->obj.symbol->model);
      if (size < 0)
	ERREXIT(1);
      symbol->value->obj.number = size;
      symbol->flags |= SYMBOL_FLAG_CONSTVALUE;
      symbol->model = model_create(C_TYPE_INT);
      syntax->obj.symbol = symbol;
      break;
    case C_TYPE_AND:
    case C_TYPE_MUL:
    case C_TYPE_ADD:
    case C_TYPE_SUB:
    case C_TYPE_INC:
    case C_TYPE_DEC:
    case C_TYPE_INV:
    case C_TYPE_NOT:
      if (op1(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
      break;

    case C_TYPE_SYMBOL:
      if ((syntax->obj.symbol->value != NULL) &&
	  (syntax->obj.symbol->value->type == C_TYPE_CAST)) {
	nentry = objentry_get_next(entry);
	if (objlist_is_end(words, nentry))
	  break;
	nsyntax = objentry_get_syntax(nentry);
	if (nsyntax->type != C_TYPE_SYMBOL)
	  break;
	objlist_extract(words, nentry);
	syntax->obj.symbol->value->obj.op.arg[0] = nsyntax->obj.symbol;
      }
      break;

    default:
      break;
    }
  }
 
  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_MUL) ||
	(syntax->type == C_TYPE_DIV) ||
	(syntax->type == C_TYPE_MOD)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_ADD) ||
	(syntax->type == C_TYPE_SUB)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_RSHIFT) ||
	(syntax->type == C_TYPE_LSHIFT)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_GTEQ) ||
	(syntax->type == C_TYPE_LTEQ) ||
	(syntax->type == C_TYPE_GT) ||
	(syntax->type == C_TYPE_LT)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_EQEQ) ||
	(syntax->type == C_TYPE_NEQ)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_AND) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_XOR) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_OR) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_ANDAND) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_OROR) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  count = 0;
  s = NULL;
  for (entry = objlist_get_tail(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_prev(entry)) { /* right to left */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_COLON) {
      if (count == 0)
	s = syntax;
      count++;
    } else if (syntax->type == C_TYPE_QUESTION) {
      if (count == 0)
	ERREXIT(1);
      if (--count == 0) {
	nentry = objentry_get_next(&syntax->entry);
	pentry = objentry_get_prev(&s->entry);
	if (objlist_is_end(words, nentry) || objlist_is_end(words, pentry) ||
	    (nentry == &s->entry) || (pentry == &syntax->entry))
	  ERREXIT(1);
	w = objlist_extract_list(words, nentry, pentry);
	symbol = _value_read(w, defines_list, symbols, static_area, stack_area);
	if (symbol == NULL)
	  ERREXIT(1);
	s->type = C_TYPE_SYMBOL;
	s->obj.symbol = symbol;
	if (op3(syntax->type, syntax, words, defines_list, symbols,
		static_area, stack_area) < 0)
	  ERREXIT(1);
	s = NULL;
      }
    }
  }

  for (entry = objlist_get_tail(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_prev(entry)) { /* right to left */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if ((syntax->type == C_TYPE_EQ) ||
	(syntax->type == C_TYPE_ADDEQ) ||
	(syntax->type == C_TYPE_SUBEQ) ||
	(syntax->type == C_TYPE_MULEQ) ||
	(syntax->type == C_TYPE_DIVEQ) ||
	(syntax->type == C_TYPE_MODEQ) ||
	(syntax->type == C_TYPE_ANDEQ) ||
	(syntax->type == C_TYPE_XOREQ) ||
	(syntax->type == C_TYPE_OREQ) ||
	(syntax->type == C_TYPE_RSHIFTEQ) ||
	(syntax->type == C_TYPE_LSHIFTEQ)) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  for (entry = objlist_get_head(words);
       !objlist_is_end(words, entry);
       entry = objentry_get_next(entry)) { /* left to right */
    syntax = objentry_get_syntax(entry);
    saveline(syntax->line);
    if (syntax->type == C_TYPE_COMMA) {
      if (op2(syntax->type, syntax, words, defines_list, symbols,
	      static_area, stack_area) < 0)
	ERREXIT(1);
    }
  }

  syntax = objlist_extract_syntax_head(words);
  if (syntax == NULL) {
    symbol = NULL;
  } else {
    if (syntax->type != C_TYPE_SYMBOL)
      ERREXITAT(syntax->line, 1);
    symbol = syntax->obj.symbol;
    if ((s = objlist_extract_syntax_head(words)) != NULL)
      ERREXITAT(s->line, 1);
  }

  return symbol;
}

static int search_term_default(objentry_t entry, int arg, void *argp)
{
  syntax_t syntax;
  int i;

  syntax = objentry_get_syntax(entry);

  if (argp == NULL) {
    if (syntax->type == arg)
      return 1;
  } else {
    for (i = 0; i < arg; i++) {
      if (syntax->type == ((int *)argp)[i])
	return 1;
    }
  }

  return 0;
}

symbol_t value_read(objlist_t words,
		    objlist_t defines_list, objlist_t symbols,
		    area_t static_area, area_t stack_area,
		    objentry_search_func_t search_func, int arg, void *argp,
		    int permit_noterm, int keep_term)
{
  objlist_t w;
  objentry_t entry;

  if (search_func == NULL) {
    if ((arg == 0) && (argp == NULL)) {
      w = objlist_extract_list_all(words);
      goto ret;
    }
    search_func = search_term_default;
  }

  entry = objlist_search_from_head(words, search_func, arg, argp);
  if (entry == NULL) {
    if (!permit_noterm)
      ERREXIT(1);
    w = objlist_extract_list_all(words);
  } else {
    w = objlist_extract_list_prev(words, entry);
    if (!keep_term) {
      objlist_extract(words, entry);
      syntax_destroy(objentry_get_syntax(entry));
    }
  }

ret:
  return _value_read(w, defines_list, symbols, static_area, stack_area);
}
